#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                $Date$
#                $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import os
import time
import threading
import traceback
import gurux_common.io.BaudRate
import gurux_common.io.Parity
import gurux_common.io.StopBits
from gurux_common.enums import TraceLevel, MediaState, TraceTypes
from gurux_common.IGXMedia import IGXMedia
from gurux_common.MediaStateEventArgs import MediaStateEventArgs
from gurux_common.TraceEventArgs import TraceEventArgs
from gurux_common.PropertyChangedEventArgs import PropertyChangedEventArgs
from gurux_common.ReceiveEventArgs import ReceiveEventArgs
from ._GXSynchronousMediaBase import _GXSynchronousMediaBase
if os.name == 'nt':  # sys.platform == 'win32':
    from ._handlers.GXWindowsHandler import GXWindowsHandler
elif os.name == 'posix':
    from ._handlers.GXLinuxHandler import GXLinuxHandler

# pylint: disable=too-many-public-methods, too-many-instance-attributes,
# too-many-arguments
class GXSerial(IGXMedia):
    def __init__(self,
                 port,
                 baudRate=gurux_common.io.BaudRate.BAUD_RATE_9600,
                 dataBits=8,
                 parity=gurux_common.io.Parity.NONE,
                 stopBits=gurux_common.io.StopBits.ONE):
        """Constructor.
        port : Serial port.
        baudRate : Baud rate.
        dataBits : Data bits.
        parity : Parity.
        stopBits : Stop bits.
        """
        self.__receiveDelay = 0
        self.__asyncWaitTime = 0
        ###Handle to serial port handler.
        self.__h = self.__initialize()
        #Values are saved if port is not open and user try to set them.
        #Serial port baud rate.
        self.__h.baudRate = baudRate
        ###Used data bits.
        self.__h.dataBits = dataBits
        ###Stop bits.
        self.__h.stopBits = stopBits
        ###Used parity.
        self.__h.parity = parity
        self.__portName = port
        self.__syncBase = _GXSynchronousMediaBase(100)
        ###Amount of bytes sent.
        self.__bytesSent = 0
        self.__bytesReceived = 0
        ###Trace level.
        self.__trace = TraceLevel.OFF
        ###End of packet.
        self.__eop = None
        self.__listeners = []
        self.__rtsEnable = False
        self.__handshake = None
        self.__receiver = None
        self.__closing = threading.Event()
        self.__lock = threading.Lock()

    #pylint: disable=unused-private-member
    def __getTrace(self):
        return self.__trace

    #pylint: disable=unused-private-member
    def __setTrace(self, value):
        self.__trace = value
        self.__syncBase.trace = value

    trace = property(__getTrace, __setTrace)
    """Trace level."""

    def addListener(self, listener):
        self.__listeners.append(listener)

    def removeListener(self, listener):
        self.__listeners.remove(listener)

    def __notifyPropertyChanged(self, info):
        """Notify that property has changed."""
        for it in self.__listeners:
            it.onPropertyChanged(self, PropertyChangedEventArgs(info))

    def __notifyError(self, ex):
        """Notify clients from error occurred."""
        for it in self.__listeners:
            it.onError(self, ex)
            if self.__trace & TraceLevel.ERROR != 0:
                it.onTrace(self, TraceEventArgs(TraceTypes.ERROR, ex))

    def __notifyReceived(self, e):
        """Notify clients from new data received."""
        for it in self.__listeners:
            it.onReceived(self, e)

    def __notifyTrace(self, e):
        """Notify clients from trace events."""
        for it in self.__listeners:
            it.onTrace(self, e)

    @classmethod
    def __initialize(cls):
        ###Initialize Gurux serial port library.
        if os.name == 'nt':
            return GXWindowsHandler()
        if os.name == 'posix':
            return GXLinuxHandler()
        raise ImportError("Sorry: no implementation for your platform ('{}') available".format(os.name))

    @classmethod
    def getPortNames(cls):
        """Gets an array of serial port names for the current computer."""
        return cls.__initialize().getPortNames()

    def getAvailableBaudRates(self):
        # pylint: disable=no-self-use
        """Get baud rates supported by given serial port."""
        return (gurux_common.io.BaudRate.BAUD_RATE_300, gurux_common.io.BaudRate.BAUD_RATE_600,
                gurux_common.io.BaudRate.BAUD_RATE_1800, gurux_common.io.BaudRate.BAUD_RATE_2400,
                gurux_common.io.BaudRate.BAUD_RATE_4800, gurux_common.io.BaudRate.BAUD_RATE_9600,
                gurux_common.io.BaudRate.BAUD_RATE_19200, gurux_common.io.BaudRate.BAUD_RATE_38400)

    def send(self, data, receiver=None):
        if not self.__h:
            raise Exception("Serial port is not open.")
        if self.__trace == TraceLevel.VERBOSE:
            self.__notifyTrace(TraceEventArgs(TraceTypes.SENT, data))

        #Reset last position if end of packet is used.
        with self.__syncBase.getSync():
            self.__syncBase.resetLastPosition()

        if not isinstance(data, bytes):
            data = bytes(_GXSynchronousMediaBase.toBytes(data))
        self.__h.write(data)
        self.__bytesSent += len(data)

    def __notifyMediaStateChange(self, state):
        ###Notify client from media state change.
        for it in self.__listeners:
            if self.__trace & TraceLevel.ERROR != 0:
                it.onTrace(self, TraceEventArgs(TraceTypes.INFO, state))
            it.onMediaStateChange(self, MediaStateEventArgs(state))

    #Handle received data.
    def __handleReceivedData(self, buff, info):
        self.__bytesReceived += len(buff)
        totalCount = 0
        if self.getIsSynchronous():
            arg = None
            with self.__syncBase.getSync():
                self.__syncBase.appendData(buff, 0, len(buff))
                #Search end of packet if it is given.
                if self.eop:
                    tmp = _GXSynchronousMediaBase.toBytes(self.eop)
                    totalCount = _GXSynchronousMediaBase.indexOf(buff, tmp, 0, len(buff))
                if totalCount != -1:
                    if self.trace == TraceLevel.VERBOSE:
                        arg = TraceEventArgs(TraceTypes.RECEIVED, buff, 0, totalCount + 1)
                    self.__syncBase.setReceived()
            if arg:
                self.__notifyTrace(arg)
        else:
            self.__syncBase.resetReceivedSize()
            if self.trace == TraceLevel.VERBOSE:
                self.__notifyTrace(TraceEventArgs(TraceTypes.RECEIVED, buff))
            e = ReceiveEventArgs(buff, info)
            self.__notifyReceived(e)

    def __readThread(self):
        #pylint: disable=broad-except, bare-except
        while not self.__closing.isSet():
            try:
                data = self.__h.read()
                if data is not None:
                    self.__handleReceivedData(data, self.__portName)
                    #Give some time before read next bytes.  In this way we are
                    #not reading data one byte at the time.
                    time.sleep(0.1)
            except Exception as ex:
                #If serial port is removed.
                if not self.isOpen():
                    self.__closing.set()
                    self.__notifyMediaStateChange(MediaState.CLOSED)
                    self.__bytesSent = 0
                    self.__syncBase.exception = ex
                    self.__syncBase.resetReceivedSize()
                    self.__syncBase.setReceived()
                if not self.__closing.isSet():
                    traceback.print_exc()

    def open(self):
        self.close()
        if not self.__portName:
            raise Exception("Serial port is not selected.")

        with self.__syncBase.getSync():
            self.__syncBase.resetLastPosition()

        self.__notifyMediaStateChange(MediaState.OPENING)
        if self.__trace & TraceLevel.INFO != 0:
            eopString = str(self.eop)
            self.__notifyTrace(TraceEventArgs(TraceTypes.INFO,\
                "Settings: Port: " + self.__portName + " Baud Rate: " + str(self.baudRate) + \
                " Data Bits: " + str(int(self.dataBits)) + " Parity: " + str(self.parity) + \
               " Stop Bits: " + str(self.stopBits) + " Eop:" + eopString))

        self.__h.open(self.__portName)
        self.__closing.clear()
        self.rtsEnable = True
        self.dtrEnable = True
        self.__notifyMediaStateChange(MediaState.OPEN)
        self.__receiver = threading.Thread(target=self.__readThread)
        self.__receiver.start()

    def close(self):
        if self.__receiver:
            self.__closing.set()
            self.__notifyMediaStateChange(MediaState.CLOSING)
            self.__h.close()
            if self.__receiver:
                self.__receiver.join()
                self.__receiver = None
            self.__notifyMediaStateChange(MediaState.CLOSED)
            self.__bytesSent = 0
            self.__syncBase.resetReceivedSize()

    def __getBaudRate(self):
        if self.__h.isOpen():
            self.__h.baudRate = gurux_common.io.BaudRate(self.__h.getBaudRate())
        return self.__h.baudRate

    def __setBaudRate(self, value):
        if self.__h.baudRate != value:
            self.__h.baudRate = value
            if self.__h.isOpen():
                self.__h.setBaudRate(int(value))
            self.__notifyPropertyChanged("baudRate")

    baudRate = property(__getBaudRate, __setBaudRate)
    """baud rate."""

    def setBreakState(self, value):
        """Set break state."""
        self.__h.setBreakState(value)

    def getBytesToRead(self):
        """Gets the number of bytes in the receive buffer."""
        return self.__h.getBytesToRead()

    def getBytesToWrite(self):
        """Gets the number of bytes in the send buffer."""
        return self.__h.getBytesToWrite()

    def getCDHolding(self):
        """Gets the state of the Carrier Detect line for the port."""
        return self.__h.getCDHolding()

    def getCtsHolding(self):
        """Gets the state of the Clear-to-Send line."""
        return self.__h.getCtsHolding()

    def __getDataBits(self):
        if self.__h.isOpen():
            self.__h.dataBits = self.__h.getDataBits()
        return self.__h.dataBits

    def __setDataBits(self, value):
        self.__h.dataBits = value
        if self.__h.isOpen():
            self.__h.setDataBits(int(value))

    dataBits = property(__getDataBits, __setDataBits)
    """The standard length of data bits per byte."""

    def getDsrHolding(self):
        """"Gets the state of the Data Set Ready (DSR) signal."""
        return self.__h.getDsrHolding()

    #pylint: disable=unused-private-member
    def __getDtrEnable(self):
        return self.__h.getDtrEnable()

    #pylint: disable=unused-private-member
    def __setDtrEnable(self, value):
        self.__h.setDtrEnable(value)

    dtrEnable = property(__getDtrEnable, __setDtrEnable)
    """Is Data Terminal Ready (DTR) signal enabled."""

    def __getHandshake(self):
        if self.__h.isOpen():
            self.__handshake = self.__h.getHandshake()
        return self.__handshake

    def __setHandshake(self, value):
        self.__handshake = value

    handshake = property(__getHandshake, __setHandshake)
    """Handshaking protocol for serial port transmission of data."""

    def isOpen(self):
        return self.__h and self.__h.isOpen()

    def __getParity(self):
        if self.__h.isOpen():
            self.__h.parity = gurux_common.io.Parity(self.__h.getParity())
        return self.__h.parity

    def __setParity(self, value):
        self.__h.parity = value
        if self.__h.isOpen():
            self.__h.setParity(int(self.__h.parity))

    parity = property(__getParity, __setParity)
    """Parity-checking protocol."""

    def __getPortName(self):
        return self.__portName

    def __setPortName(self, value):
        if value != self.__portName:
            self.__portName = value
            self.__notifyPropertyChanged("port")

    port = property(__getPortName, __setPortName)
    """The port for communications, including but not limited to all available COM ports."""

    def __getRtsEnable(self):
        return self.__rtsEnable

    def __setRtsEnable(self, value):
        self.__rtsEnable = value

    rtsEnable = property(__getRtsEnable, __setRtsEnable)
    """Is Request to Send (RTS) signal enabled during serial communication."""

    def __getStopBits(self):
        if self.__h.isOpen():
            self.__h.stopBits = gurux_common.io.StopBits(self.__h.getStopBits())
        return self.__h.stopBits

    def __setStopBits(self, value):
        self.__h.stopBits = value
        if self.__h.isOpen():
            self.__h.setStopBits(int(self.__h.stopBits))

    stopBits = property(__getStopBits, __setStopBits)
    """Sets the standard number of stop bits per byte."""

    def receive(self, args):
        return self.__syncBase.receive(args)

    def getBytesSent(self):
        """Sent byte count."""
        return self.__bytesSent

    def getBytesReceived(self):
        """Received byte count."""
        return self.__bytesReceived

    def resetByteCounters(self):
        """Resets BytesReceived and BytesSent counters."""
        self.__bytesSent = 0
        self.__bytesReceived = 0

    def getSettings(self):
        """Media settings as a XML string."""
        sb = ""
        nl = "\r\n"
        if self.__portName:
            sb += "<Port>"
            sb += self.__portName
            sb += "</Port>"
            sb += nl

        if self.__h.baudRate != gurux_common.io.BaudRate.BAUD_RATE_9600:
            sb += "<BaudRate>"
            sb += str(int(self.__h.baudRate))
            sb += "</BaudRate>"
            sb += nl

        if self.__h.stopBits != gurux_common.io.StopBits.ONE:
            sb += "<StopBits>"
            sb += str(int(self.__h.stopBits))
            sb += "</StopBits>"
            sb += nl

        if self.__h.parity != gurux_common.io.Parity.NONE:
            sb += "<Parity>"
            sb += str(int(self.__h.parity))
            sb += "</Parity>"
            sb += nl

        if self.__h.dataBits != 8:
            sb += "<DataBits>"
            sb += str(int(self.__h.dataBits))
            sb += "</DataBits>"
            sb += nl
        return sb

    def setSettings(self, value):
        #Reset to default values.
        self.__portName = ""
        self.__h.baudRate = gurux_common.io.BaudRate.BAUD_RATE_9600
        self.__h.stopBits = gurux_common.io.StopBits.ONE
        self.__h.parity = gurux_common.io.Parity.NONE
        self.__h.dataBits = 8

    def copy(self, target):
        self.__portName = target.portName
        self.__h.baudRate = target.baudRate
        self.__h.stopBits = target.stopBits
        self.__h.parity = target.parity
        self.__h.dataBits = target.dataBits

    def getName(self):
        return self.__portName

    def getMediaType(self):
        return "Serial"

    def getSynchronous(self):
        return self.__lock

    #pylint: disable=no-member
    def getIsSynchronous(self):
        return self.__lock.locked()

    def resetSynchronousBuffer(self):
        with self.__syncBase.getSync():
            self.__syncBase.resetReceivedSize()

    def validate(self):
        if not self.__portName:
            raise Exception("Invalid port name.")

    def __getEop(self):
        return self.__eop

    def __setEop(self, value):
        self.__eop = value

    eop = property(__getEop, __setEop)

    def getReceiveDelay(self):
        return self.__receiveDelay

    def setReceiveDelay(self, value):
        self.__receiveDelay = value

    def getAsyncWaitTime(self):
        return self.__asyncWaitTime

    def setAsyncWaitTime(self, value):
        self.__asyncWaitTime = value

    def __str__(self):
        return self.__portName + ":" + str(int(self.__h.baudRate)) + " " +\
            str(self.__h.dataBits) + str(self.__h.parity.name) + str(int(self.__h.stopBits) + 1)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                $Date$
#                $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import threading
import datetime
from gurux_common.enums.TraceLevel import TraceLevel

###Python 2 requires this
#pylint: disable=bad-option-value,old-style-class
class _GXSynchronousMediaBase:
    def __init__(self, length=0):
        """Constructor.

        length : receive buffer size.
        """
        #Trace level.
        self.trace = TraceLevel.OFF
        #Occurred exception.
        self.exception = None
        #Received bytes.
        self.__receivedBuffer = bytearray(length)
        #Received event.
        self.__receivedEvent = threading.Event()
        #Synchronous object.
        self.__sync = threading.RLock()
        #Amount of received bytes.
        self.__receivedSize = 0
        #Position where end of packet was last time search.  This is used to
        #improve searching.
        self.__lastPosition = 0

    def resetLastPosition(self):
        """Reset last position."""
        self.__lastPosition = 0

    def resetReceivedSize(self):
        """Reset received size."""
        self.__receivedSize = 0

    def setReceived(self):
        """Set received event."""
        self.__receivedEvent.set()

    def getSync(self):
        """Get synchronous object."""
        return self.__sync

    def getReceivedSize(self):
        """@return Amount of received bytes."""
        return self.__receivedSize

    def getReceivedData(self):
        """@return Get received data."""
        return self.__receivedBuffer[0:self.__receivedSize]

    def appendData(self, data, index, count):
        """Append new data.

        data : data to append.
        index : Index where start.
        count : Count of bytes to add.
        """
        with self.__sync:
            #Allocate new buffer.
            if self.__receivedSize + count > len(self.__receivedBuffer):
                tmp = self.__receivedBuffer
                self.__receivedBuffer = bytearray(2 * len(self.__receivedBuffer))
                self.__receivedBuffer[0:self.__receivedSize] = tmp[0:self.__receivedSize]

            self.__receivedBuffer[self.__receivedSize:self.__receivedSize + count - index] = data[index: count - index]
            self.__receivedSize += count - index

    @classmethod
    def indexOf(cls, data, pattern, index, count):
        """
        Finds the first occurrence of the pattern in the text.
        data : Data where to find.
        pattern : Search pattern.
        index : Byte index to start.
        count : Count of bytes to search.
        Return True if pattern is found.
        """
        failure = _GXSynchronousMediaBase.__computeFailure(pattern)
        j = 0
        if not data or len(data) < index:
            return -1

        for i in range(index, count):
            while j > 0 and pattern[j] != data[i]:
                j = failure[j - 1]
            if pattern[j] == data[i]:
                j = j + 1

            if j == len(pattern):
                return i - len(pattern) + 1
        return -1

    @classmethod
    def __computeFailure(cls, pattern):
        """Computes the failure function using a boot-strapping process, where the
        pattern is matched against itself.

        pattern : Pattern to search.
        Returns Failure pattern.
        """
        failure = bytearray(len(pattern))
        j = 0
        for i in range(1, len(pattern)):
            while j > 0 and pattern[j] != pattern[i]:
                j = failure[j - 1]
            if pattern[j] == pattern[i]:
                j = j + 1

            failure[i] = j
        return failure

    def __findData(self, args):
        """Find data from buffer.
        args : Receive parameters.
        isFound : Is data found in given time.
        Returns : Position where end of packet was found. -1 Is returned if data
        was not found in given time.
        """
        nSize = 0
        foundPosition = -1
        lastBuffSize = 0
        startTime = datetime.datetime.now().time().microsecond
        terminator = None
        nMinSize = args.count
        if nMinSize < nSize:
            nMinSize = nSize

        waitTime = args.waitTime
        self.exception = None
        if waitTime <= 0:
            waitTime = -1
        if args.eop:
            terminator = self.toBytes(args.eop)
            nSize = len(terminator)
        #Wait until reply occurred.
        while foundPosition == -1:
            if waitTime == 0:
                #If we want to read all data.
                if args.allData:
                    foundPosition = self.__receivedSize
                else:
                    foundPosition = -1
                break

            if waitTime != - 1:
                waitTime = args.waitTime - ((datetime.datetime.now().time().microsecond - startTime) / 1000)
                if waitTime < 0:
                    waitTime = 0

            with self.__sync:
                isReceived = not (lastBuffSize == self.__receivedSize or self.__receivedSize < nMinSize)

            # Do not wait if there is data on the buffer...
            if not isReceived:
                if waitTime == - 1:
                    isReceived = self.__receivedEvent.wait()
                elif waitTime != 0:
                    isReceived = self.__receivedEvent.wait(waitTime / 1000)
                self.__receivedEvent.clear()
            if self.exception:
                raise self.exception
            #If timeout occurred.
            if not isReceived:
                #If we want to read all data.
                if args.allData and self.__receivedSize != 0:
                    foundPosition = self.__receivedSize
                else:
                    foundPosition = -1
                break

            with self.__sync:
                lastBuffSize = self.__receivedSize
                #Read more data, if not enough.
                if self.__receivedSize < nMinSize:
                    continue

                #If only byte count matters.
                if nSize == 0:
                    foundPosition = args.count
                else:
                    if self.__lastPosition != 0 and self.__lastPosition < self.__receivedSize:
                        index = self.__lastPosition
                    else:
                        index = args.count
                    #If terminator found.
                    if isinstance(args.eop, (list)):
                        raise Exception("Only single byte terminator allowed.")
                    if len(terminator) != 1 and self.__receivedSize - index < len(terminator):
                        index = self.__receivedSize - len(terminator)
                    foundPosition = self.indexOf(self.__receivedBuffer, terminator, index, self.__receivedSize)
                    self.__lastPosition = self.__receivedSize
                    if foundPosition != -1:
                        foundPosition += len(terminator)

        #If terminator is not given read only bytes that are needed.
        if nSize == 0:
            foundPosition = args.count
        return foundPosition

    def receive(self, args):
        """
        Receive new data synchronously from the media.
        args : Receive parameters.
        Return True if new data is received.
        """
        if args.eop is None and args.count == 0:
            raise Exception("Either Count or Eop must be set.")
        foundPosition = self.__findData(args)
        data = None
        if foundPosition != -1:
            with self.__sync:
                if args.allData:
                    #If all data is copied.
                    foundPosition = self.__receivedSize

                if foundPosition != 0:
                    #Convert bytes to object.
                    data = self.__receivedBuffer[0:foundPosition]
                    #Remove read data.
                    self.__receivedSize -= foundPosition
                    #Received size can go less than zero if we have received
                    #data and we try to read more.
                    if self.__receivedSize < 0:
                        self.__receivedSize = 0
                    if self.__receivedSize != 0:
                        self.__receivedBuffer[0:self.__receivedSize] = self.__receivedBuffer[foundPosition:foundPosition + self.__receivedSize]
            #Reset count after read.
            args.count = 0
            #Append data.
            oldReplySize = 0
            if args.reply is None:
                args.reply = data
            else:
                oldArray = args.reply
                newArray = data
                if newArray:
                    oldReplySize = len(oldArray)
                    len_ = oldReplySize + len(newArray)
                    arr = bytearray(len_)
                    #Copy old values.
                    arr[0:len(oldArray)] = args.reply[0:len(oldArray)]
                    #Copy new values.
                    arr[len(oldArray): len(oldArray) + len(newArray)] = newArray[0:len(newArray)]
                    args.reply = arr
        return foundPosition != -1

    #Convert value to bytes.
    @classmethod
    def toBytes(cls, value):
        if isinstance(value, bytearray):
            pass
        elif isinstance(value, str):
            value = bytearray(value.encode())
        elif isinstance(value, (bytes, list)):
            value = bytearray(value)
        elif isinstance(value, memoryview):
            value = bytearray(value.tobytes())
        elif isinstance(value, int):
            value = bytearray([value])
        else:
            raise ValueError("Invalid data value.")
        return value
